###################################################################################################
# A tutorial on using the "bridge" function from the networktools package
###################################################################################################

## Install the networktools package
install.packages("networktools")
require("networktools")

## Load your network
## This can be an adjacency matrix, a qgraph object, or an igraph object

## Here we will use data from Jones, Mair, Mugno, Riemann, & McNally (2017)
adolescent_OCD <- read.csv("Data_for_tutorial.csv")[c(14:29, 31:40)]

## Let's rename the columns to be more friendly for interpretation:
colnames(adolescent_OCD)<- c("onset", "middle", "late", "hypersom", "sad", "decappetite", "incappetite", "weightloss", "weightgain", "concen", "guilt", "suicide", "anhedonia", "fatigue", "retard", "agitation", "obtime", "obinterfer", "obdistress", "obresist", "obcontrol", "comptime", "compinterf", "compdis", "compresis", "compcont")

## This contains data on 87 adolescents with primary OCD, who were assessed for OCD and MDD symptoms
## The first 16 columns are depression symptoms, the next 10 are OCD symptoms
dim(adolescent_OCD)
head(adolescent_OCD)

## Now let's use the qgraph package to create an EBICglasso network:
require(qgraph)
my_network <- EBICglasso(cor(adolescent_OCD), n=87, gamma=0.3)
my_network ## This is an adjacency matrix!

## In this case, the disorders each represent a community:
## Let's make a vector to keep our "community structure" straight
community_structure <- c(rep("Depression", 16), rep("OCD", 10))
community_structure

## Let's plot our network with qgraph:
qgraph(my_network, groups=community_structure, layout="spring", color=c("lightgreen", "orange"))

## Now we are ready to look at bridge centrality!
## It's a one-liner with the bridge() function:

bridge_centrality <- bridge(my_network, communities=community_structure)
bridge_centrality

## It's very easy to get a nice plot of this:
plot(bridge_centrality) ## We are missing a couple of values because a couple nodes are completely unconnected

## We can narrow down the type of plots to display using the "include" argument
plot(bridge_centrality, include="Bridge Strength")

## We can also change how the nodes are ordered in the plot, using the "order" argument
plot(bridge_centrality, include=c("Bridge Strength", "Bridge Betweenness"), order="value")

## You can also plot z-values instead of raw centrality scores:
plot(bridge_centrality, include=c("Bridge Strength", "Bridge Betweenness"), zscore=TRUE)

## We can extract each type of bridge centrality easily:
bridge_centrality$`Bridge Strength`
bridge_centrality$`Bridge Betweenness`
bridge_centrality$`Bridge Closeness`
bridge_centrality$`Bridge Expected Influence (1-step)`
bridge_centrality$`Bridge Expected Influence (2-step)`

## What about those nice plots where bridges are colored?
## It's a little more tricky, but we can do it

## Select the top 80th percentile bridge strength:
bridge_strength <- bridge_centrality$`Bridge Strength`
top_bridges <- names(bridge_strength[bridge_strength>quantile(bridge_strength, probs=0.80, na.rm=TRUE)])

## Now create a new community vector where bridges are their own "community"
bridge_num_w1 <- which(names(bridge_strength) %in% top_bridges)
new_communities <- vector()
for(i in 1:length(bridge_strength)) {
  if(i %in% bridge_num_w1) {
    new_communities[i] <- "Bridge"
  } else {new_communities[i] <- community_structure[i]}
}

## And now use that community vector as your "groups" in qgraph! 
qgraph(my_network, layout="spring", groups=new_communities, color=c("lightblue", "lightgreen", "orange"))

## Check out ?bridge and ?plot.bridge for extra information. Good luck!



